% Boosted MaxEnt classifier
%
% X           : N-by-D feature matrix with N samples and D dimensions
% P_gt        : N-by-K distribution matrix with K-label distribution for each sample
%                 (values must be in range [0, 1])
% T           : number of iterations to train for
% step_size   : step_size / sqrt(t)
% regressor_type : string specified type of regressor to train
% subsample : [0,1] fraction indicating random percentage to subsample at
%               each iteration, neg value indicate use all data
%
% model   : struct containing 
%             .regressors T-by-(K-1) cell array with regressors
%             .steps T-by-1 vector with step sizes
function [model] = bme_train(X, P_gt, T, step_size, regressor_type, subsample)
assert(nargin == 5 || nargin == 6);
if (nargin == 5)
    subsample = -1;
end
if (nargin == 6)
    assert(subsample <= 1.0);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
MIN_RESIDUAL = 0.015;   % ignore fgs with small residual (diff in distribution)
MIN_PERCENT_LABELED = 0.75;

nbr_samples = size(X,1);
nbr_labels = size(P_gt, 2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sanity checks
assert(nbr_samples == size(P_gt,1));
% verify distributions contain probabilities
% (values be positive and each row should sum to <= 1)
assert(sum(min(P_gt) < 0.0) == 0);
assert(sum( sum(P_gt,2) > (ones(nbr_samples,1)+1e-4) ) == 0);

ignore_indices = P_gt < 0.01 & repmat(sum(P_gt,2) < MIN_PERCENT_LABELED, 1, nbr_labels);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
steps = zeros(T, 1);
regressors = cell(T, 1);
for t=1:T
    t
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Predict with current model
    Q = bme_predict(regressors, steps, regressor_type, X, nbr_labels, t-1);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Plot losses (assume predictions for partially labeled samples are true)
    P_gt(ignore_indices) = Q(ignore_indices);
    losses = sum(-P_gt .* log(Q), 2);
    plot(losses, 'k.');
    title(num2str(t));
    drawnow;
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Train regressor
    fg_residuals = P_gt - Q;
    if (subsample > 0.0)
        rnd_order = randperm(nbr_samples)';
        rnd_order = rnd_order(1:round(subsample*nbr_samples));
        fg_residuals = fg_residuals(rnd_order,:);
        X_t = X(rnd_order,:);
        regressors{t} = regressor_train(regressor_type, X_t, fg_residuals, MIN_RESIDUAL);
    else
        regressors{t} = regressor_train(regressor_type, X, fg_residuals, MIN_RESIDUAL);
    end
    steps(t) = step_size / sqrt(t);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Save model during progress
    if (0 && mod(t,5) == 0)
        fn_out = ['iter_' , num2str(t) , '.mat'];
        save(fn_out,  'regressors','steps');
    end
end

model.regressors = regressors;
model.steps = steps;
model.nbr_labels = nbr_labels;
model.regressor_type = regressor_type;
end
