function [ X_normalized, means, stds ] = normalize_data(X, means, stds)
% Normalizes data (zero mean and unit variance along each feature dimension)
%
% Inputs
% -------
% X     : N-by-D feature matrix with N samples that have D dimensions
% means : (Optional) 1-by-D vector containing means to use to normalize X.
%                    If not provided, will be computed from X
% stds  : (Optional) 1-by-D vector containing the std deviations to use to
%                    normalize X. If not provided, will be computed from X
%
% Outputs
% -------
% X_normalized : The normalized versioned of X
% means        : 1-by-D vector w/ the means used to normalize X.  Hence, if 
%                this method was called with all parameters, then this is
%                the same as the passed-in means vector
% stds         : 1-by-D vector w/ the std dev. used to normalize X.  Hence, if 
%                this method was called with all parameters, then this is
%                the same as the passed-in stds vector

N = size(X, 1);
D = size(X, 2);

if (nargin == 1)
    means = mean(X);
    stds = std(X);
end

assert(length(means) == D);
assert(length(stds) == D);

% Zero mean
X_normalized = X - repmat(means, N, 1);

% Make unit variance for dimensions that actually vary
for d=1:D
    if (stds(d) > eps)
        X_normalized(:,d) = X_normalized(:,d) ./ stds(d);
    else
        X_normalized(:,d) = zeros(N, 1);
    end
end

X_normalized(X_normalized > 5) = 5;
X_normalized(X_normalized < -5) = -5;
end

