% Compute label context feature for each segment
%
% label_conf_maps : K-by-1 cell vector for each label, containing per-pixel label confidences
% seg_map         : M-by-N segment map with values [0, S-1]
%
% segs_label_context : S-by-1 cell vector containing 1-by-2K vector w/ above and below confidences per segment
function segs_label_context = create_label_context_feature(label_conf_maps, seg_map)
DILATE_FACTOR = 30;

% Ensure image dimensions are the same
assert(size(seg_map,1) == size(label_conf_maps{1},1));
assert(size(seg_map,2) == size(label_conf_maps{1},2));

nbr_segs = max(max(seg_map)) + 1;
nbr_labels = size(label_conf_maps,1);

% For each segment, compute label context for area above and below
segs_label_context = cell(nbr_segs,1);
for i=1:nbr_segs
    [seg_rows, seg_cols] = find(seg_map == (i-1));
    ctr_row = round(mean(seg_rows));
    
    % Compute mask around current segment
    curr_seg_mask = seg_map == (i-1);
    outer_seg_mask = xor(bwmorph(curr_seg_mask, 'dilate', DILATE_FACTOR), curr_seg_mask);
    
    % Separate into above and below masks
    above_seg_mask = outer_seg_mask;
    above_seg_mask(ctr_row:end,:) = 0;
    below_seg_mask = outer_seg_mask;
    below_seg_mask(1:ctr_row,:) = 0;
    
    segs_label_context{i} = zeros(1, 2*nbr_labels);
    % Average area above center
    for k=1:nbr_labels
        segs_label_context{i}(k) = mean(label_conf_maps{k}(above_seg_mask));
    end
    % Average area below center
    for k=1:nbr_labels
        segs_label_context{i}(nbr_labels+k) = mean(label_conf_maps{k}(below_seg_mask));
    end
    nan_indices = isnan(segs_label_context{i});
    segs_label_context{i}(nan_indices) = 0;
end
