% Creates a hierarchical tree of segmentations from the output of 
% Arbelaez et al.'s UCM stuff
%
% ucm2           : 2M-by-2N the output from contours2ucm(gPb_orient, 'doubleSize')
% vec_scales : Vector of scales in decreasing order
%
% seg_tree : Q-by-1 cell vector, where Q is the number of scales, containing 
%              structs with the fields:
%                  .scale - the scale in the hiearchy (ucm2 <= .scale)
%                  .curr_seg_map - M-by-N matrix containing values [0,S1-1]
%                                  for each pixel's segment, where S1 is 
%                                  the number of segments in current level
%                  .parent_seg_list - 1-by-S1 vector with values of [0,S2-1]
%                                      for each current segment's parent 
%                                      segment in seg_tree{i-1}, where S2 
%                                      is the number of segments in level i-1
%                                      (all -1 for top layer seg_tree{1})
function seg_tree = create_ucm_tree(ucm2, vec_scales)

disp('warning: passing vector of scales instead of scale interval');

nbr_scales = length(vec_scales);
assert(nbr_scales > 1);
seg_tree = cell(nbr_scales, 1);

nbr_img_rows = floor(size(ucm2,1) / 2);
nbr_img_cols = floor(size(ucm2,2) / 2);
prev_seg_map = -1 * ones(nbr_img_rows, nbr_img_cols); % image parent is undefined
for i=1:length(vec_scales)
    curr_scale = vec_scales(i);

    %%%%%%%%%%%%%%%%%%
    % get the partition at current scale
    partitions2 = bwlabel(ucm2 <= curr_scale);
    % put in original scale of image
    % (range of values: [0,S], where 0 = boundary)
    partitions = partitions2(2:2:end, 2:2:end);
    
    %%%%%%%%%%%%%%%%%%
    % iterate over reach segment
    %   -create current level segment mask
    %   -create parent segment mask
    curr_seg_map = zeros(nbr_img_rows, nbr_img_cols);
    nbr_segments = max(max(partitions));
    parent_seg_list = -ones(1, nbr_segments);
    for j=1:nbr_segments
        curr_seg_mask = partitions == j;
        
        % Populate the segmentation map with current index (start from 0)
        curr_seg_map(curr_seg_mask) = j-1;
        
        % Examine parent segment index
        % (sub-segment comes from only 1 parent segment)
        parent_seg_index = unique(prev_seg_map(curr_seg_mask));
        assert(length(parent_seg_index) == 1);
        parent_seg_list(j) = parent_seg_index;
    end
    
    seg_tree{i,1}.scale = curr_scale;
    seg_tree{i,1}.curr_seg_map = curr_seg_map;
    seg_tree{i,1}.parent_seg_list = parent_seg_list;
    
    prev_seg_map = curr_seg_map;
end
