% Loads all features for the image
%
% path_seg_tree_root : The path to the directory that contains scale_XXX subdirs
%                        used when creating segmentation tree
% path_inf_tree_root : The path to directory that contains scale_XXX subdirs
%                        to load parent label predictions 
% img_basename  : basename of the image
% feature_names : cell vector containing feature names to load
% curr_scale    : current segment scale
% parent_scale  : parent segment scale (-1 to indicate top layer)
%
% S_x         : S-by-D matrix containing feature vectors for each segment
% S_y_par_inf : either: S-by-K matrix containing respective segment's parent 
%                 predictions for K labels, or empty if no parent
% S_y_gt      : S-by-K matrix containing the segment's ground truth labels
% parent_ilp  : either: 1-by-K vector with ILP using parent level predictions,
%                 or empty if no parent/top level parent
% S_lcontext  : S-by-2K matrix containing label context feature for each
%                 segment, computed using parent, or empty if no parent/top level parent
function [S_x, S_y_par_inf, S_y_gt, parent_ilp, S_lcontext, seg_map] = ...
    load_img_segs2(path_seg_tree_root, path_inf_tree_root, img_basename, feature_names, curr_scale, parent_scale)
assert(nargin == 6);

path_seg_tree_curr_scale = [path_seg_tree_root, '/scale_', num2str(curr_scale)];
path_inf_tree_parent_scale = [path_inf_tree_root, '/scale_', num2str(parent_scale)];
                 
%%%%%%%%%%%%%%%%%%%%%
% Load segment map for current image
seg_map = load([path_seg_tree_curr_scale, '/segment_maps/', img_basename , '.seg']);
assert(min(min(seg_map)) == 0);
nbr_segs = max(max(seg_map)) + 1;

%%%%%%%%%%%%%%%%%%%%%
% Compute features from parent predictions:
S_y_par_inf = [];
S_lcontext = [];
parent_ilp = [];
if (parent_scale >= 0)
    % %%%%%%%%%%%%%%%%%%%% Parent predictions %%%%%%%%%%%%%%%%%%%%%%%
    % Load parent segment list for current level
    parent_seg_list = load([path_seg_tree_curr_scale, '/parent_', num2str(parent_scale), '_lists/', img_basename, '_parent.list']);
    assert(min(parent_seg_list) == 0);
    assert(length(parent_seg_list) == nbr_segs);
    nbr_par_segs = max(parent_seg_list) + 1;
    % Load parent level's predicted histogram
    par_inf_histos = load([path_inf_tree_parent_scale, '/inf_segment_labels/', img_basename, '.histo_labels']);
    assert(size(par_inf_histos,1) == nbr_par_segs);
    % Align current level segment's with parent segment predictions
    S_y_par_inf = par_inf_histos(parent_seg_list+1, :);
    
    %%%%%%%%%%%%%%%%%%%%%
    % 1) Compute image-level prior from parent predictions
    % 2) Compute label-context histogram using parent predictions
    if (parent_scale < 256)
        % %%%%%%%%%%%%%%%%%%%% Image level prior %%%%%%%%%%%%%%%%%%%%%%%
        % Load parent segment map
        parent_seg_map = load([path_seg_tree_root, '/scale_', num2str(parent_scale), '/segment_maps/', img_basename , '.seg']);
        parent_ilp = create_image_level_prior(parent_seg_map, par_inf_histos);
        
        % %%%%%%%%%%%%%%%%%%%% Label context %%%%%%%%%%%%%%%%%%%%%%%
        % Label context for current segments using parent predictions
        parent_label_conf_maps = create_label_conf_maps(parent_seg_map, par_inf_histos);
        S_lcontext = create_label_context_feature(parent_label_conf_maps, seg_map);
        S_lcontext = cell2mat(S_lcontext);
    end
end

%%%%%%%%%%%%%%%%%%%%%
% Load ground truth label histogram
S_y_gt = load([path_seg_tree_curr_scale, '/gt_segment_labels/', img_basename , '.histo_labels']);

%%%%%%%%%%%%%%%%%%%%%
% Load tree-independent features
S_x= [];
nbr_features = length(feature_names);
for j=1:nbr_features 
    feat_name = feature_names{j};
    X_j = load([path_seg_tree_curr_scale, '/features/', feat_name, '/', img_basename, '.features']);
    assert(size(X_j, 1) == nbr_segs);
    S_x = [S_x, X_j];
end

end
