% Trains a level in the hierarchy using stacking
%
% path_seg_tree_root
% path_machine_root  : root directory of machine to store this module 
% curr_scale    : scale in the tree that training for
% parent_scale  : parent scale
% feature_names : directory names that contain the features to use
% img_basenames : training basenames
% train_params  : .T, .step_size
% nbr_held_out  : # of images to hold out to test on for each fold (-1 to
%                   skip the folding process but no predictions saved)
%
% ASSUMING NEVER DISCARD SEGMENT
function train_stacked_level(path_seg_tree_root, path_machine_root, curr_scale, ...
            parent_scale, feature_names, img_basenames, train_params, nbr_held_out)
assert(nargin == 8);

%%%%%%%%%%%%%%%%%%%%%%%
% Create directory to store model, predicted labels for current level/module
path_curr_module = [path_machine_root, '/scale_', num2str(curr_scale)];
path_inf_segment_labels = [path_curr_module, '/inf_segment_labels'];
path_inf_conf_maps = [path_curr_module, '/inf_conf_maps'];
mkdir(path_curr_module);
mkdir(path_inf_segment_labels);
mkdir(path_inf_conf_maps);

%%%%%%%%%%%%%%%%%%%%%%%
% Calculate number of splits for stacking
nbr_imgs = length(img_basenames);
nbr_folds = ceil(nbr_imgs / nbr_held_out);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Load all features and labels
X_seg_feats = cell(nbr_imgs, 1);
X_parent_histos = cell(nbr_imgs, 1); % holds predicted labels for parent segment
X_ilp = cell(nbr_imgs,1);            % image level prior from parent predictions
X_label_context = cell(nbr_imgs,1);  % label context per segment
y_gt = cell(nbr_imgs, 1);
all_seg_maps = cell(nbr_imgs, 1);
matlabpool open
parfor i=1:nbr_imgs
    i
    curr_basename = char(img_basenames(i));
    [S_x, S_y_par_inf, S_y_gt, parent_ilp, S_lcontext, all_seg_maps{i}] = ... 
        load_img_segs2(path_seg_tree_root, path_machine_root, curr_basename, feature_names, curr_scale, parent_scale);
    nbr_segs = size(S_x,1);    

    X_seg_feats{i} = S_x;
    X_parent_histos{i} = S_y_par_inf;
    X_ilp{i} = repmat(parent_ilp, nbr_segs, 1);
    X_label_context{i} = S_lcontext;
    y_gt{i} = S_y_gt;

    assert(nbr_segs == size(S_y_gt,1));
    if (parent_scale >= 0)
        assert(nbr_segs == size(S_y_par_inf,1));
    else
        assert(isempty(S_y_par_inf));
    end
    if (parent_scale >= 0 && parent_scale < 256)
        assert(nbr_segs == size(S_lcontext,1));
    else
        assert(isempty(S_lcontext));
        assert(isempty(parent_ilp));
    end
end
matlabpool close force

%%%%%%%%%%%%%%%%%%%%
[X_seg_feats_normalize, means, stds] = normalize_data(cell2mat(X_seg_feats));
save([path_curr_module,'/normalize.mat'], 'means', 'stds');
nbr_total_segs = size(X_seg_feats_normalize,1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Start stage 1

% Make predictions from xfold trained models
X_vec_stage1 = cell(4,1);
X_vec_stage1{1} = X_seg_feats;
X_vec_stage1{2} = X_parent_histos;
X_vec_stage1{3} = X_ilp;
X_vec_stage1{4} = X_label_context;
y_infer1 = train_xfold(X_vec_stage1, y_gt, nbr_folds, nbr_held_out, train_params);

% Train classifier for stage 1
disp('Starting to train  Stage 1 ALL...');
X_stage1 = [X_seg_feats_normalize, cell2mat(X_vec_stage1{2}), cell2mat(X_vec_stage1{3}), cell2mat(X_vec_stage1{4}), ones(nbr_total_segs,1)];
clear X_vec_stage1
if (strcmp(train_params.regressor_type,'ExpMaxEnt'))
    flr_model_stage1 = eme_train(X_stage1, cell2mat(y_gt), train_params.step_size, ...
        train_params.eme_reg, train_params.eme_thresh, train_params.T);
else
    flr_model_stage1 = bme_train(X_stage1, cell2mat(y_gt), train_params.T, ...
        train_params.step_size, train_params.regressor_type, train_params.subsample);
end
disp('Finished training stage 1');

% Save model
save([path_curr_module,'/flr_model_stage1.mat'], 'flr_model_stage1');
clear flr_model_stage1

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Skip stage 2 if at the top layer
if (parent_scale < 0)
    disp('Writing predictions to file...');
    for i=1:nbr_imgs
        save_predictions2(char(img_basenames{i}), y_infer1{i}, path_inf_segment_labels);
    end
    return;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Start stage 2

% Train predictions from xvalid models
% Create label context features using predictions from stage 1
X_vec_stage2 = cell(5,1);
X_vec_stage2{1} = X_seg_feats;
X_vec_stage2{2} = X_parent_histos;
X_vec_stage2{3} = X_ilp;
X_vec_stage2{4} = y_infer1;
X_vec_stage2{5} = create_label_contexts(all_seg_maps, y_infer1);
y_infer2 = train_xfold(X_vec_stage2, y_gt, nbr_folds, nbr_held_out, train_params);

% Save predictions to pass to next layer
disp('Writing predictions to file...');
for i=1:nbr_imgs
    save_predictions2(char(img_basenames{i}), y_infer2{i}, path_inf_segment_labels);
end

% Train classifier for stage 2
disp('Starting to train Stage 2 ALL...');
X_stage2 = [X_seg_feats_normalize, cell2mat(X_vec_stage2{2}), cell2mat(X_vec_stage2{3}), cell2mat(X_vec_stage2{4}), cell2mat(X_vec_stage2{5}), ones(nbr_total_segs,1)];
clear X_vec_stage2
if (strcmp(train_params.regressor_type,'ExpMaxEnt'))
    flr_model_stage2 = eme_train(X_stage2, cell2mat(y_gt), train_params.step_size, ...
        train_params.eme_reg, train_params.eme_thresh, train_params.T);
else
    flr_model_stage2 = bme_train(X_stage2, cell2mat(y_gt), train_params.T, ...
        train_params.step_size, train_params.regressor_type, train_params.subsample);
end

% Save model
save([path_curr_module,'/flr_model_stage2.mat'], 'flr_model_stage2');
clear flr_model_stage2

end % end train_module_folds




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Make predictions on hold out datasets by training on the other part
function y_inf = train_xfold(X_vec, y_gt, nbr_folds, nbr_held_out, train_params)
nbr_features = length(X_vec);
nbr_imgs = size(X_vec{1},1);
nbr_labels = size(y_gt{1}, 2);
for i=1:nbr_features
    assert(nbr_imgs == size(X_vec{i},1));
end
assert(nbr_imgs == size(y_gt,1));

%%%%%%%%%%%%%%%%%%%%%%%
% Determine the indices for training and testing for each fold
rnd_order = randperm(nbr_imgs);
fold_train_indices = cell(nbr_folds, 1);
fold_test_indices = cell(nbr_folds, 1);
for i=1:nbr_folds
    test_start_idx = (i-1) * nbr_held_out + 1;
    test_end_idx = min(i*nbr_held_out, nbr_imgs);
    fold_test_indices{i} = rnd_order(test_start_idx:test_end_idx);
    fold_train_indices{i} = rnd_order([1:test_start_idx-1 , test_end_idx+1:nbr_imgs]);
    assert(isempty(intersect(fold_train_indices{i}, fold_test_indices{i})));
end

%%%%%%%%%%%%%%%%%%%%%%%
% preallocate to hold predicted labels so can be stored in parallel
y_inf_folds = cell(nbr_folds,1);
for i=1:nbr_folds
    y_inf_folds = cell(nbr_imgs,1); % this will be sparse
end

%%%%%%%%%%%%%%%%%%%%%%%
% Train each fold
T = train_params.T;
step_size = train_params.step_size;
regressor_t = train_params.regressor_type;
subsample = train_params.subsample;
if isfield(train_params, 'eme_reg'); eme_reg = train_params.eme_reg; else; eme_reg = -Inf; end
if isfield(train_params, 'eme_thresh'); eme_thresh = train_params.eme_thresh; else; eme_thresh = -Inf; end
%matlabpool open
for i=1:nbr_folds
    curr_fold = i
    test_indices = fold_test_indices{i};
    train_indices = fold_train_indices{i};

    %%%%%%%%%%%%%%%%%%%%%%%
    % White first feature, append the rest and add bias
    [X_train , means, stds] = normalize_data(cell2mat(X_vec{1}(train_indices)));
    for j=2:nbr_features
        X_train = [X_train , cell2mat(X_vec{j}(train_indices))];
    end
    X_train = [X_train, ones(size(X_train,1), 1)];
    y_train = cell2mat(y_gt(train_indices));
        
    %%%%%%%%%%%%%%%%%%%%%%%
    % Train and regressor
    disp('Starting to train...');
    if (strcmp(regressor_t,'ExpMaxEnt'))
        flr_model = eme_train(X_train, y_train, step_size, eme_reg, eme_thresh, T);
    else
        flr_model = bme_train(X_train, y_train, T, step_size, regressor_t, subsample);
    end
    X_train = [];
    y_train = [];
    
    %%%%%%%%%%%%%%%%%%%%%%%
    % Output predictions over segments
    disp('Making predictions....');
    for te_idx=test_indices
        % White first feature, append the rest, and add bias
        X_test = normalize_data(X_vec{1}{te_idx}, means, stds);
        for j=2:nbr_features
            X_test = [X_test , X_vec{j}{te_idx}];
        end
        X_test = [X_test, ones(size(X_test,1),1)];

        % Infer
        if (strcmp(regressor_t,'ExpMaxEnt'))
            y_test = eme_predict(flr_model, X_test);
        else
            y_test = bme_predict(flr_model.regressors, flr_model.steps, regressor_t, X_test, nbr_labels);
        end
        
        % Need to do this for matlab's parfor
        y_inf_folds{i}{te_idx} = y_test;
        
        X_test = [];
        y_test = [];
    end
    flr_model = [];
    disp('------ end fold ------');
end
%matlabpool close force

%%%%%%%%%%%%%%%%%%%%%%%
% Put predicted responses in 1...N image order
y_inf = cell(nbr_imgs,1);
for i=1:nbr_folds
    test_indices = fold_test_indices{i};
    for te_idx = test_indices
        y_inf{te_idx} = y_inf_folds{i}{te_idx};
    end
end
end




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Given predictions per image, create label-context feature for each segment
function img_label_contexts = create_label_contexts(all_seg_maps, img_seg_labels)
nbr_imgs = size(all_seg_maps,1);
assert(nbr_imgs == size(img_seg_labels,1));

disp('Creating label context feature...');
img_label_contexts = cell(nbr_imgs,1);
matlabpool open
parfor i=1:nbr_imgs
    curr_seg_map = all_seg_maps{i};
    curr_seg_labels = img_seg_labels{i}; % labels per segmap
    assert(size(curr_seg_labels,1) == (max(max(curr_seg_map))+1));

    % Create confidence maps for each label
    label_conf_maps = create_label_conf_maps(curr_seg_map, curr_seg_labels);
    
    % Create label context features around each segment
    img_label_contexts{i} = cell2mat(create_label_context_feature(label_conf_maps, curr_seg_map));
end
matlabpool close force
end
